!----------------------------------------------------------------------!
!                          S P A R S K I T                             !
!----------------------------------------------------------------------!
!                     UNARY SUBROUTINES MODULE                         !
!----------------------------------------------------------------------!
! contents:                                                            !
!----------                                                            !
! submat : extracts a submatrix from a sparse matrix.                  !
! filter : filters elements from a matrix according to their magnitude.!
! filterm: same as above, but for the MSR format                       !
! csort  : sorts the elements in increasing order of columns           !
! clncsr : clean up the CSR format matrix, remove duplicate entry, etc !
! transp : in-place transposition routine (see also csrcsc in formats) !
! copmat : copy of a matrix into another matrix (both stored csr)      !
! msrcop : copies a matrix in MSR format into a matrix in MSR format   !
! getelm : returns a(i,j) for any (i,j) from a CSR-stored matrix.      !
! getdia : extracts a specified diagonal from a matrix.                !
! getl   : extracts lower triangular part                              !
! getu   : extracts upper triangular part                              !
! levels : gets the level scheduling structure for lower triangular    !
!          matrices.                                                   !
! amask  : extracts     C = A mask M                                   !
! rperm  : permutes the rows of a matrix (B = P A)                     !
! cperm  : permutes the columns of a matrix (B = A Q)                  !
! dperm  : permutes both the rows and columns of a matrix (B = P A Q ) !
! dperm1 : general extractiob routine (extracts arbitrary rows)        !
! dperm2 : general submatrix permutation/extraction routine            !
! dmperm : symmetric permutation of row and column (B=PAP') in MSR fmt !
! dvperm : permutes a real vector (in-place)                           !
! ivperm : permutes an integer vector (in-place)                       !
! retmx  : returns the max absolute value in each row of the matrix    !
! diapos : returns the positions of the diagonal elements in A.        !
! extbdg : extracts the main diagonal blocks of a matrix.              !
! getbwd : returns the bandwidth information on a matrix.              !
! blkfnd : finds the block-size of a matrix.                           !
! blkchk : checks whether a given integer is the block size of A.      !
! infdia : obtains information on the diagonals of A.                  !
! amubdg : gets number of nonzeros in each row of A*B (as well as NNZ) ! 
! aplbdg : gets number of nonzeros in each row of A+B (as well as NNZ) !
! rnrms  : computes the norms of the rows of A                         !
! cnrms  : computes the norms of the columns of A                      !
! roscal : scales the rows of a matrix by their norms.                 !
! coscal : scales the columns of a matrix by their norms.              !
! addblk : Adds a matrix B into a block of A.                          !
! get1up : Collects the first elements of each row of the upper        !
!          triangular portion of the matrix.                           !
! xtrows : extracts given rows from a matrix in CSR format.            !
! csrkvstr:  Finds block row partitioning of matrix in CSR format      !
! csrkvstc:  Finds block column partitioning of matrix in CSR format   !
! kvstmerge: Merges block partitionings, for conformal row/col pattern !
!----------------------------------------------------------------------!
      subroutine submat (n,job,i1,i2,j1,j2,a,ja,ia,nr,nc,ao,jao,iao)
      integer n,job,i1,i2,j1,j2,nr,nc,ia(*),ja(*),jao(*),iao(*)
      real*8 a(*),ao(*) 
!-----------------------------------------------------------------------
! extracts the submatrix A(i1:i2,j1:j2) and puts the result in 
! matrix ao,iao,jao
!---- In place: ao,jao,iao may be the same as a,ja,ia.
!-------------- 
! on input
!---------
! n	= row dimension of the matrix 
! i1,i2 = two integers with i2 .ge. i1 indicating the range of rows to be
!          extracted. 
! j1,j2 = two integers with j2 .ge. j1 indicating the range of columns 
!         to be extracted.
!         * There is no checking whether the input values for i1, i2, j1,
!           j2 are between 1 and n. 
! a,
! ja,
! ia    = matrix in compressed sparse row format. 
!
! job	= job indicator: if job .ne. 1 then the real values in a are NOT
!         extracted, only the column indices (i.e. data structure) are.
!         otherwise values as well as column indices are extracted...
!         
! on output
!-------------- 
! nr	= number of rows of submatrix 
! nc	= number of columns of submatrix 
!	  * if either of nr or nc is nonpositive the code will quit.
!
! ao,
! jao,iao = extracted matrix in general sparse format with jao containing
!	the column indices,and iao being the pointer to the beginning 
!	of the row,in arrays a,ja.
!----------------------------------------------------------------------!
!           Y. Saad, Sep. 21 1989                                      !
!----------------------------------------------------------------------!
      nr = i2-i1+1
      nc = j2-j1+1
!     
      if ( nr .le. 0 .or. nc .le. 0) return
!     
      klen = 0
!     
!     simple procedure. proceeds row-wise...
!     
      do 100 i = 1,nr
         ii = i1+i-1
         k1 = ia(ii)
         k2 = ia(ii+1)-1
         iao(i) = klen+1
!-----------------------------------------------------------------------
         do 60 k=k1,k2
            j = ja(k)
            if (j .ge. j1 .and. j .le. j2) then
               klen = klen+1
               if (job .eq. 1) ao(klen) = a(k)
               jao(klen) = j - j1+1
            endif
 60      continue
 100  continue
      iao(nr+1) = klen+1
      return
!------------end-of submat---------------------------------------------- 
!----------------------------------------------------------------------- 
      end
!-----------------------------------------------------------------------
      subroutine filter(n,job,drptol,a,ja,ia,b,jb,ib,len,ierr)
      real*8 a(*),b(*),drptol
      integer ja(*),jb(*),ia(*),ib(*),n,job,len,ierr
!-----------------------------------------------------------------------
!     This module removes any elements whose absolute value
!     is small from an input matrix A and puts the resulting
!     matrix in B.  The input parameter job selects a definition
!     of small.
!-----------------------------------------------------------------------
! on entry:
!---------
!  n	 = integer. row dimension of matrix
!  job   = integer. used to determine strategy chosen by caller to
!         drop elements from matrix A. 
!          job = 1  
!              Elements whose absolute value is less than the
!              drop tolerance are removed.
!          job = 2
!              Elements whose absolute value is less than the 
!              product of the drop tolerance and the Euclidean
!              norm of the row are removed. 
!          job = 3
!              Elements whose absolute value is less that the
!              product of the drop tolerance and the largest
!              element in the row are removed.
! 
! drptol = real. drop tolerance used for dropping strategy.
! a	
! ja
! ia     = input matrix in compressed sparse format
! len	 = integer. the amount of space available in arrays b and jb.
!
! on return:
!---------- 
! b	
! jb
! ib    = resulting matrix in compressed sparse format. 
! 
! ierr	= integer. containing error message.
!         ierr .eq. 0 indicates normal return
!         ierr .gt. 0 indicates that there is'nt enough
!         space is a and ja to store the resulting matrix.
!         ierr then contains the row number where filter stopped.
! note:
!------ This module is in place. (b,jb,ib can ne the same as 
!       a, ja, ia in which case the result will be overwritten).
!----------------------------------------------------------------------!
!           contributed by David Day,  Sep 19, 1989.                   !
!----------------------------------------------------------------------!
! local variables
      real*8 norm,loctol
      integer index,row,k,k1,k2 
!
      index = 1
      do 10 row= 1,n
         k1 = ia(row)
         k2 = ia(row+1) - 1
         ib(row) = index
	   goto (100,200,300) job
 100     norm = 1.0d0
         goto 400
 200     norm = 0.0d0
         do 22 k = k1,k2
            norm = norm + a(k) * a(k)
 22      continue
         norm = sqrt(norm)
         goto 400
 300     norm = 0.0d0
         do 23 k = k1,k2
            if( abs(a(k))  .gt. norm) then
               norm = abs(a(k))
            endif
 23      continue
 400     loctol = drptol * norm
	   do 30 k = k1,k2
	    if( abs(a(k)) .gt. loctol)then 
               if (index .gt. len) then
               ierr = row 
               return
            endif
            b(index) =  a(k)
            jb(index) = ja(k)
            index = index + 1
         endif
 30   continue
 10   continue
      ib(n+1) = index
      return
!--------------------end-of-filter -------------------------------------
!-----------------------------------------------------------------------
      end
!-----------------------------------------------------------------------
      subroutine filterm (n,job,drop,a,ja,b,jb,len,ierr)
      real*8 a(*),b(*),drop
      integer ja(*),jb(*),n,job,len,ierr
!-----------------------------------------------------------------------
!     This subroutine removes any elements whose absolute value
!     is small from an input matrix A. Same as filter but
!     uses the MSR format.
!-----------------------------------------------------------------------
! on entry:
!---------
!  n	   = integer. row dimension of matrix
!  job   = integer. used to determine strategy chosen by caller to
!         drop elements from matrix A. 
!          job = 1  
!              Elements whose absolute value is less than the
!              drop tolerance are removed.
!          job = 2
!              Elements whose absolute value is less than the 
!              product of the drop tolerance and the Euclidean
!              norm of the row are removed. 
!          job = 3
!              Elements whose absolute value is less that the
!              product of the drop tolerance and the largest
!              element in the row are removed.
! 
! drop = real. drop tolerance used for dropping strategy.
! a	
! ja     = input matrix in Modifief Sparse Row format
! len	   = integer. the amount of space in arrays b and jb.
!
! on return:
!---------- 
!
! b, jb = resulting matrix in Modifief Sparse Row format
! 
! ierr	= integer. containing error message.
!         ierr .eq. 0 indicates normal return
!         ierr .gt. 0 indicates that there is'nt enough
!         space is a and ja to store the resulting matrix.
!         ierr then contains the row number where filter stopped.
! note:
!------ This module is in place. (b,jb can ne the same as 
!       a, ja in which case the result will be overwritten).
!----------------------------------------------------------------------!
!           contributed by David Day,  Sep 19, 1989.                   !
!----------------------------------------------------------------------!
! local variables
!
      real*8 norm,loctol
      integer index,row,k,k1,k2 
!
      index = n+2
      do 10 row= 1,n
         k1 = ja(row)
         k2 = ja(row+1) - 1
         jb(row) = index
	   goto (100,200,300) job
 100     norm = 1.0d0
         goto 400
 200     norm = a(row)**2 
         do 22 k = k1,k2
            norm = norm + a(k) * a(k)
 22      continue
         norm = sqrt(norm)
         goto 400
 300     norm = abs(a(row)) 
         do 23 k = k1,k2
            norm = max(abs(a(k)),norm) 
 23      continue
 400     loctol = drop * norm
	   do 30 k = k1,k2
	    if( abs(a(k)) .gt. loctol)then 
               if (index .gt. len) then
                  ierr = row 
                  return
               endif
               b(index) =  a(k)
               jb(index) = ja(k)
               index = index + 1
            endif
 30      continue
 10   continue
      jb(n+1) = index
      return
!--------------------end-of-filterm-------------------------------------
!-----------------------------------------------------------------------
      end
!-----------------------------------------------------------------------
      subroutine csort (n,a,ja,ia,iwork,values) 
      logical values
      integer n, ja(*), ia(n+1), iwork(*) 
      real*8 a(*) 
!-----------------------------------------------------------------------
! This routine sorts the elements of  a matrix (stored in Compressed
! Sparse Row Format) in increasing order of their column indices within 
! each row. It uses a form of bucket sort with a cost of O(nnz) where
! nnz = number of nonzero elements. 
! requires an integer work array of length 2*nnz.  
!-----------------------------------------------------------------------
! on entry:
!--------- 
! n     = the row dimension of the matrix
! a     = the matrix A in compressed sparse row format.
! ja    = the array of column indices of the elements in array a.
! ia    = the array of pointers to the rows. 
! iwork = integer work array of length max ( n+1, 2*nnz ) 
!         where nnz = (ia(n+1)-ia(1))  ) .
! values= logical indicating whether or not the real values a(*) must 
!         also be permuted. if (.not. values) then the array a is not
!         touched by csort and can be a dummy array. 
! 
! on return:
!----------
! the matrix stored in the structure a, ja, ia is permuted in such a
! way that the column indices are in increasing order within each row.
! iwork(1:nnz) contains the permutation used  to rearrange the elements.
!----------------------------------------------------------------------- 
! Y. Saad - Feb. 1, 1991.
!-----------------------------------------------------------------------
! local variables
      integer i, k, j, ifirst, nnz, next  
!
! count the number of elements in each column
!
      do 1 i=1,n+1
         iwork(i) = 0
 1    continue
      do 3 i=1, n
         do 2 k=ia(i), ia(i+1)-1 
            j = ja(k)+1
            iwork(j) = iwork(j)+1
 2       continue 
 3    continue
!
! compute pointers from lengths. 
!
      iwork(1) = 1
      do 4 i=1,n
         iwork(i+1) = iwork(i) + iwork(i+1)
 4    continue
! 
! get the positions of the nonzero elements in order of columns.
!
      ifirst = ia(1) 
      nnz = ia(n+1)-ifirst
      do 5 i=1,n
         do 51 k=ia(i),ia(i+1)-1 
            j = ja(k) 
            next = iwork(j) 
            iwork(nnz+next) = k
            iwork(j) = next+1
 51      continue
 5    continue
!
! convert to coordinate format
! 
      do 6 i=1, n
         do 61 k=ia(i), ia(i+1)-1 
            iwork(k) = i
 61      continue
 6    continue
!
! loop to find permutation: for each element find the correct 
! position in (sorted) arrays a, ja. Record this in iwork. 
! 
      do 7 k=1, nnz
         ko = iwork(nnz+k) 
         irow = iwork(ko)
         next = ia(irow)
!
! the current element should go in next position in row. iwork
! records this position. 
! 
         iwork(ko) = next
         ia(irow)  = next+1
 7       continue
!
! perform an in-place permutation of the  arrays.
! 
         call ivperm (nnz, ja(ifirst), iwork) 
         if (values) call dvperm (nnz, a(ifirst), iwork) 
!
! reshift the pointers of the original matrix back.
! 
      do 8 i=n,1,-1
         ia(i+1) = ia(i)
 8    continue
      ia(1) = ifirst 
!
      return 
!---------------end-of-csort-------------------------------------------- 
!-----------------------------------------------------------------------
      end
!-----------------------------------------------------------------------
      subroutine clncsr(job,value2,nrow,a,ja,ia,indu,iwk)
!     .. Scalar Arguments ..
      integer job, nrow, value2
!     ..
!     .. Array Arguments ..
      integer ia(nrow+1),indu(nrow),iwk(nrow+1),ja(*)
      real*8  a(*)
!     ..
!
!     This routine performs two tasks to clean up a CSR matrix
!     -- remove duplicate/zero entries,
!     -- perform a partial ordering, new order lower triangular part,
!        main diagonal, upper triangular part.
!
!     On entry:
!
!     job   = options
!         0 -- nothing is done
!         1 -- eliminate duplicate entries, zero entries.
!         2 -- eliminate duplicate entries and perform partial ordering.
!         3 -- eliminate duplicate entries, sort the entries in the
!              increasing order of clumn indices.
!
!     value2  -- 0 the matrix is pattern only (a is not touched)
!                1 matrix has values too.
!     nrow    -- row dimension of the matrix
!     a,ja,ia -- input matrix in CSR format
!
!     On return:
!     a,ja,ia -- cleaned matrix
!     indu    -- pointers to the beginning of the upper triangular
!                portion if job > 1
!
!     Work space:
!     iwk     -- integer work space of size nrow+1
!
!     .. Local Scalars ..
      integer i,j,k,ko,ipos,kfirst,klast
      real*8  tmp
!     ..
!
      if (job.le.0) return
!
!     .. eliminate duplicate entries --
!     array INDU is used as marker for existing indices, it is also the
!     location of the entry.
!     IWK is used to stored the old IA array.
!     matrix is copied to squeeze out the space taken by the duplicated
!     entries.
!
      do 90 i = 1, nrow
         indu(i) = 0
         iwk(i) = ia(i)
 90   continue
      iwk(nrow+1) = ia(nrow+1)
      k = 1
      do 120 i = 1, nrow
         ia(i) = k
         ipos = iwk(i)
         klast = iwk(i+1)
 100     if (ipos.lt.klast) then
            j = ja(ipos)
            if (indu(j).eq.0) then
!     .. new entry ..
               if (value2.ne.0) then
                  if (a(ipos) .ne. 0.0D0) then
                     indu(j) = k
                     ja(k) = ja(ipos)
                     a(k) = a(ipos)
                     k = k + 1
                  endif
               else
                  indu(j) = k
                  ja(k) = ja(ipos)
                  k = k + 1
               endif
            else if (value2.ne.0) then
!     .. duplicate entry ..
               a(indu(j)) = a(indu(j)) + a(ipos)
            endif
            ipos = ipos + 1
            go to 100
         endif
!     .. remove marks before working on the next row ..
         do 110 ipos = ia(i), k - 1
            indu(ja(ipos)) = 0
 110     continue
 120  continue
      ia(nrow+1) = k
      if (job.le.1) return
!
!     .. partial ordering ..
!     split the matrix into strict upper/lower triangular
!     parts, INDU points to the the beginning of the upper part.
!
      do 140 i = 1, nrow
         klast = ia(i+1) - 1
         kfirst = ia(i)
 130     if (klast.gt.kfirst) then
            if (ja(klast).lt.i .and. ja(kfirst).ge.i) then
!     .. swap klast with kfirst ..
               j = ja(klast)
               ja(klast) = ja(kfirst)
               ja(kfirst) = j
               if (value2.ne.0) then
                  tmp = a(klast)
                  a(klast) = a(kfirst)
                  a(kfirst) = tmp
               endif
            endif
            if (ja(klast).ge.i)                                           &
     &         klast = klast - 1
            if (ja(kfirst).lt.i)                                          &
     &         kfirst = kfirst + 1
            go to 130
         endif
!
         if (ja(klast).lt.i) then
            indu(i) = klast + 1
         else
            indu(i) = klast
         endif
 140  continue
      if (job.le.2) return
!
!     .. order the entries according to column indices
!     burble-sort is used
!
      do 190 i = 1, nrow
         do 160 ipos = ia(i), indu(i)-1
            do 150 j = indu(i)-1, ipos+1, -1
               k = j - 1
               if (ja(k).gt.ja(j)) then
                  ko = ja(k)
                  ja(k) = ja(j)
                  ja(j) = ko
                  if (value2.ne.0) then
                     tmp = a(k)
                     a(k) = a(j)
                     a(j) = tmp
                  endif
               endif
 150        continue
 160     continue
         do 180 ipos = indu(i), ia(i+1)-1
            do 170 j = ia(i+1)-1, ipos+1, -1
               k = j - 1
               if (ja(k).gt.ja(j)) then
                  ko = ja(k)
                  ja(k) = ja(j)
                  ja(j) = ko
                  if (value2.ne.0) then
                     tmp = a(k)
                     a(k) = a(j)
                     a(j) = tmp
                  endif
               endif
 170        continue
 180     continue
 190  continue
      return
!---- end of clncsr ----------------------------------------------------
!-----------------------------------------------------------------------
      end
!----------------------------------------------------------------------- 
      subroutine copmat (nrow,a,ja,ia,ao,jao,iao,ipos,job) 
      real*8 a(*),ao(*) 
      integer nrow, ia(*),ja(*),jao(*),iao(*), ipos, job 
!----------------------------------------------------------------------
! copies the matrix a, ja, ia, into the matrix ao, jao, iao. 
!----------------------------------------------------------------------
! on entry:
!---------
! nrow	= row dimension of the matrix 
! a,
! ja,
! ia    = input matrix in compressed sparse row format. 
! ipos  = integer. indicates the position in the array ao, jao
!         where the first element should be copied. Thus 
!         iao(1) = ipos on return. 
! job   = job indicator. if (job .ne. 1) the values are not copies 
!         (i.e., pattern only is copied in the form of arrays ja, ia).
!
! on return:
!----------
! ao,
! jao,
! iao   = output matrix containing the same data as a, ja, ia.
!-----------------------------------------------------------------------
!           Y. Saad, March 1990. 
!-----------------------------------------------------------------------
! local variables
      integer kst, i, k 
!
      kst    = ipos -ia(1) 
      do 100 i = 1, nrow+1
         iao(i) = ia(i) + kst
 100  continue
!     
      do 200 k=ia(1), ia(nrow+1)-1
         jao(kst+k)= ja(k)
 200  continue
!
      if (job .ne. 1) return
      do 201 k=ia(1), ia(nrow+1)-1
         ao(kst+k) = a(k)
 201  continue
!
      return
!--------end-of-copmat -------------------------------------------------
!-----------------------------------------------------------------------
      end
!-----------------------------------------------------------------------
      subroutine msrcop (nrow,a,ja,ao,jao,job) 
      real*8 a(*),ao(*) 
      integer nrow, ja(*),jao(*), job 
!----------------------------------------------------------------------
! copies the MSR matrix a, ja, into the MSR matrix ao, jao 
!----------------------------------------------------------------------
! on entry:
!---------
! nrow	= row dimension of the matrix 
! a,ja  = input matrix in Modified compressed sparse row format. 
! job   = job indicator. Values are not copied if job .ne. 1 
!       
! on return:
!----------
! ao, jao   = output matrix containing the same data as a, ja.
!-----------------------------------------------------------------------
!           Y. Saad, 
!-----------------------------------------------------------------------
! local variables
      integer i, k 
!
      do 100 i = 1, nrow+1
         jao(i) = ja(i) 
 100  continue
!     
      do 200 k=ja(1), ja(nrow+1)-1
         jao(k)= ja(k)
 200  continue
!
      if (job .ne. 1) return
      do 201 k=ja(1), ja(nrow+1)-1
         ao(k) = a(k)
 201  continue
      do 202 k=1,nrow
         ao(k) = a(k)
 202  continue
!
      return
!--------end-of-msrcop -------------------------------------------------
!-----------------------------------------------------------------------
      end
!----------------------------------------------------------------------- 
      double precision function getelm (i,j,a,ja,ia,iadd,sorted) 
!-----------------------------------------------------------------------
!     purpose:
!     -------- 
!     this function returns the element a(i,j) of a matrix a, 
!     for any pair (i,j).  the matrix is assumed to be stored 
!     in compressed sparse row (csr) format. getelm performs a
!     binary search in the case where it is known that the elements 
!     are sorted so that the column indices are in increasing order. 
!     also returns (in iadd) the address of the element a(i,j) in 
!     arrays a and ja when the search is successsful (zero if not).
!----- 
!     first contributed by noel nachtigal (mit). 
!     recoded jan. 20, 1991, by y. saad [in particular
!     added handling of the non-sorted case + the iadd output] 
!-----------------------------------------------------------------------
!     parameters:
!     ----------- 
! on entry: 
!---------- 
!     i      = the row index of the element sought (input).
!     j      = the column index of the element sought (input).
!     a      = the matrix a in compressed sparse row format (input).
!     ja     = the array of column indices (input).
!     ia     = the array of pointers to the rows' data (input).
!     sorted = logical indicating whether the matrix is knonw to 
!              have its column indices sorted in increasing order 
!              (sorted=.true.) or not (sorted=.false.).
!              (input). 
! on return:
!----------- 
!     getelm = value of a(i,j). 
!     iadd   = address of element a(i,j) in arrays a, ja if found,
!              zero if not found. (output) 
!
!     note: the inputs i and j are not checked for validity. 
!-----------------------------------------------------------------------
!     noel m. nachtigal october 28, 1990 -- youcef saad jan 20, 1991.
!----------------------------------------------------------------------- 
      integer i, ia(*), iadd, j, ja(*)
      double precision a(*)
      logical sorted 
!
!     local variables.
!
      integer ibeg, iend, imid, k
!
!     initialization 
!
      iadd = 0 
      getelm = 0.0
      ibeg = ia(i)
      iend = ia(i+1)-1
!
!     case where matrix is not necessarily sorted
!     
      if (.not. sorted) then 
!
! scan the row - exit as soon as a(i,j) is found
!
         do 5  k=ibeg, iend
            if (ja(k) .eq.  j) then
               iadd = k 
               goto 20 
            endif
 5       continue
!     
!     end unsorted case. begin sorted case
!     
      else
!     
!     begin binary search.   compute the middle index.
!     
 10      imid = ( ibeg + iend ) / 2
!     
!     test if  found
!     
         if (ja(imid).eq.j) then
            iadd = imid 
            goto 20
         endif
         if (ibeg .ge. iend) goto 20
!     
!     else     update the interval bounds. 
!     
         if (ja(imid).gt.j) then
            iend = imid -1
         else 
            ibeg = imid +1
         endif
         goto 10  
!     
!     end both cases
!     
      endif
!     
 20   if (iadd .ne. 0) getelm = a(iadd) 
!
      return
!--------end-of-getelm--------------------------------------------------
!-----------------------------------------------------------------------
      end 
!-----------------------------------------------------------------------
      subroutine getdia (nrow,ncol,job,a,ja,ia,len,diag,idiag,ioff)
      real*8 diag(*),a(*)
      integer nrow, ncol, job, len, ioff, ia(*), ja(*), idiag(*)
!-----------------------------------------------------------------------
! this subroutine extracts a given diagonal from a matrix stored in csr 
! format. the output matrix may be transformed with the diagonal removed
! from it if desired (as indicated by job.) 
!----------------------------------------------------------------------- 
! our definition of a diagonal of matrix is a vector of length nrow
! (always) which contains the elements in rows 1 to nrow of
! the matrix that are contained in the diagonal offset by ioff
! with respect to the main diagonal. if the diagonal element
! falls outside the matrix then it is defined as a zero entry.
! thus the proper definition of diag(*) with offset ioff is 
!
!     diag(i) = a(i,ioff+i) i=1,2,...,nrow
!     with elements falling outside the matrix being defined as zero.
! 
!----------------------------------------------------------------------- 
! 
! on entry:
!---------- 
!
! nrow	= integer. the row dimension of the matrix a.
! ncol	= integer. the column dimension of the matrix a.
! job   = integer. job indicator.  if job = 0 then
!         the matrix a, ja, ia, is not altered on return.
!         if job.ne.0  then getdia will remove the entries
!         collected in diag from the original matrix.
!         this is done in place.
!
! a,ja,
!    ia = matrix stored in compressed sparse row a,ja,ia,format
! ioff  = integer,containing the offset of the wanted diagonal
!	  the diagonal extracted is the one corresponding to the
!	  entries a(i,j) with j-i = ioff.
!	  thus ioff = 0 means the main diagonal
!
! on return:
!----------- 
! len   = number of nonzero elements found in diag.
!         (len .le. min(nrow,ncol-ioff)-max(1,1-ioff) + 1 )
!
! diag  = real*8 array of length nrow containing the wanted diagonal.
!	  diag contains the diagonal (a(i,j),j-i = ioff ) as defined 
!         above. 
!
! idiag = integer array of  length len, containing the poisitions 
!         in the original arrays a and ja of the diagonal elements
!         collected in diag. a zero entry in idiag(i) means that 
!         there was no entry found in row i belonging to the diagonal.
!         
! a, ja,
!    ia = if job .ne. 0 the matrix is unchanged. otherwise the nonzero
!         diagonal entries collected in diag are removed from the 
!         matrix and therefore the arrays a, ja, ia will change.
!	  (the matrix a, ja, ia will contain len fewer elements) 
! 
!----------------------------------------------------------------------!
!     Y. Saad, sep. 21 1989 - modified and retested Feb 17, 1996.      ! 
!----------------------------------------------------------------------!
!     local variables
      integer istart, max, iend, i, kold, k, kdiag, ko
!     
      istart = max(0,-ioff)
      iend = min(nrow,ncol-ioff)
      len = 0
      do 1 i=1,nrow
         idiag(i) = 0
	   diag(i) = 0.0d0 
 1    continue
!     
!     extract  diagonal elements
!     
      do 6 i=istart+1, iend
         do 51 k= ia(i),ia(i+1) -1
            if (ja(k)-i .eq. ioff) then
               diag(i)= a(k)
               idiag(i) = k
               len = len+1
               goto 6
            endif
 51      continue
 6    continue
      if (job .eq. 0 .or. len .eq.0) return
!
!     remove diagonal elements and rewind structure
! 
      ko = 0
      do  7 i=1, nrow 
         kold = ko
         kdiag = idiag(i) 
         do 71 k= ia(i), ia(i+1)-1 
            if (k .ne. kdiag) then
               ko = ko+1
               a(ko) = a(k)
               ja(ko) = ja(k)
            endif
 71      continue
         ia(i) = kold+1
 7    continue
!
!     redefine ia(nrow+1)
!
      ia(nrow+1) = ko+1
      return
!------------end-of-getdia----------------------------------------------
!-----------------------------------------------------------------------
      end
!-----------------------------------------------------------------------
      subroutine transp (nrow,ncol,a,ja,ia,iwk,ierr)
      integer nrow, ncol, ia(*), ja(*), iwk(*), ierr
      real*8 a(*) 
!------------------------------------------------------------------------
! In-place transposition routine.
!------------------------------------------------------------------------
! this subroutine transposes a matrix stored in compressed sparse row 
! format. the transposition is done in place in that the arrays a,ja,ia
! of the transpose are overwritten onto the original arrays.
!------------------------------------------------------------------------
! on entry:
!--------- 
! nrow	= integer. The row dimension of A.
! ncol	= integer. The column dimension of A.
! a	= real array of size nnz (number of nonzero elements in A).
!         containing the nonzero elements 
! ja	= integer array of length nnz containing the column positions
! 	  of the corresponding elements in a.
! ia	= integer of size n+1, where n = max(nrow,ncol). On entry
!         ia(k) contains the position in a,ja of  the beginning of 
!         the k-th row.
!
! iwk	= integer work array of same length as ja.
!
! on return:
!----------
!
! ncol	= actual row dimension of the transpose of the input matrix.
!         Note that this may be .le. the input value for ncol, in
!         case some of the last columns of the input matrix are zero
!         columns. In the case where the actual number of rows found
!         in transp(A) exceeds the input value of ncol, transp will
!         return without completing the transposition. see ierr.
! a,
! ja,
! ia	= contains the transposed matrix in compressed sparse
!         row format. The row dimension of a, ja, ia is now ncol.
!
! ierr	= integer. error message. If the number of rows for the
!         transposed matrix exceeds the input value of ncol,
!         then ierr is  set to that number and transp quits.
!         Otherwise ierr is set to 0 (normal return).
!
! Note: 
!----- 1) If you do not need the transposition to be done in place
!         it is preferrable to use the conversion routine csrcsc 
!         (see conversion routines in formats).
!      2) the entries of the output matrix are not sorted (the column
!         indices in each are not in increasing order) use csrcsc
!         if you want them sorted.
!----------------------------------------------------------------------!
!           Y. Saad, Sep. 21 1989                                      !
!  modified Oct. 11, 1989.                                             !
!----------------------------------------------------------------------!
! local variables
      real*8 t, t1
      ierr = 0
      nnz = ia(nrow+1)-1
!
!     determine column dimension
!
      jcol = 0
      do 1 k=1, nnz
         jcol = max(jcol,ja(k))
 1    continue
      if (jcol .gt. ncol) then
         ierr = jcol
         return
      endif
!     
!     convert to coordinate format. use iwk for row indices.
!     
      ncol = jcol
!     
      do 3 i=1,nrow
         do 2 k=ia(i),ia(i+1)-1
            iwk(k) = i
 2       continue 
 3    continue
!     find pointer array for transpose. 
      do 35 i=1,ncol+1
         ia(i) = 0
 35   continue
      do 4 k=1,nnz
         i = ja(k)
         ia(i+1) = ia(i+1)+1
 4    continue 
      ia(1) = 1 
!------------------------------------------------------------------------
      do 44 i=1,ncol
         ia(i+1) = ia(i) + ia(i+1)
 44   continue 
!     
!     loop for a cycle in chasing process. 
!     
      init = 1
      k = 0
 5    t = a(init)
      i = ja(init)
      j = iwk(init)
      iwk(init) = -1
!------------------------------------------------------------------------
 6    k = k+1 		
!     current row number is i.  determine  where to go. 
      l = ia(i)
!     save the chased element. 
      t1 = a(l)
      inext = ja(l)
!     then occupy its location.
      a(l)  = t
      ja(l) = j
!     update pointer information for next element to be put in row i. 
      ia(i) = l+1
!     determine  next element to be chased
      if (iwk(l) .lt. 0) goto 65
      t = t1
      i = inext
      j = iwk(l)
      iwk(l) = -1
      if (k .lt. nnz) goto 6
      goto 70
 65   init = init+1
      if (init .gt. nnz) goto 70
      if (iwk(init) .lt. 0) goto 65
!     restart chasing --	
      goto 5
 70   continue
      do 80 i=ncol,1,-1 
         ia(i+1) = ia(i)
 80   continue
      ia(1) = 1
!
      return
!------------------end-of-transp ----------------------------------------
!------------------------------------------------------------------------
      end 
!------------------------------------------------------------------------ 
      subroutine getl (n,a,ja,ia,ao,jao,iao)
      integer n, ia(*), ja(*), iao(*), jao(*)
      real*8 a(*), ao(*)
!------------------------------------------------------------------------
! this subroutine extracts the lower triangular part of a matrix 
! and writes the result ao, jao, iao. The routine is in place in
! that ao, jao, iao can be the same as a, ja, ia if desired.
!-----------
! on input:
!
! n     = dimension of the matrix a.
! a, ja, 
!    ia = matrix stored in compressed sparse row format.
! On return:
! ao, jao, 
!    iao = lower triangular matrix (lower part of a) 
!	stored in a, ja, ia, format
! note: the diagonal element is the last element in each row.
! i.e. in  a(ia(i+1)-1 ) 
! ao, jao, iao may be the same as a, ja, ia on entry -- in which case
! getl will overwrite the result on a, ja, ia.
!
!------------------------------------------------------------------------
! local variables
      real*8 t
      integer ko, kold, kdiag, k, i
!
! inititialize ko (pointer for output matrix)
!
      ko = 0
      do  7 i=1, n
         kold = ko
         kdiag = 0
         do 71 k = ia(i), ia(i+1) -1
            if (ja(k)  .gt. i) goto 71
            ko = ko+1
            ao(ko) = a(k)
            jao(ko) = ja(k)
            if (ja(k)  .eq. i) kdiag = ko
 71      continue
         if (kdiag .eq. 0 .or. kdiag .eq. ko) goto 72
!
!     exchange
!
         t = ao(kdiag)
         ao(kdiag) = ao(ko)
         ao(ko) = t
!     
         k = jao(kdiag)
         jao(kdiag) = jao(ko)
         jao(ko) = k
 72      iao(i) = kold+1
 7    continue
!     redefine iao(n+1)
      iao(n+1) = ko+1
      return
!----------end-of-getl ------------------------------------------------- 
!-----------------------------------------------------------------------
      end
!-----------------------------------------------------------------------
      subroutine getu (n,a,ja,ia,ao,jao,iao)
      integer n, ia(*), ja(*), iao(*), jao(*)
      real*8 a(*), ao(*)
!------------------------------------------------------------------------
! this subroutine extracts the upper triangular part of a matrix 
! and writes the result ao, jao, iao. The routine is in place in
! that ao, jao, iao can be the same as a, ja, ia if desired.
!-----------
! on input:
!
! n     = dimension of the matrix a.
! a, ja, 
!    ia = matrix stored in a, ja, ia, format
! On return:
! ao, jao, 
!    iao = upper triangular matrix (upper part of a) 
!	stored in compressed sparse row format
! note: the diagonal element is the last element in each row.
! i.e. in  a(ia(i+1)-1 ) 
! ao, jao, iao may be the same as a, ja, ia on entry -- in which case
! getu will overwrite the result on a, ja, ia.
!
!------------------------------------------------------------------------
! local variables
      real*8 t 
      integer ko, k, i, kdiag, kfirst 
      ko = 0
      do  7 i=1, n
         kfirst = ko+1
         kdiag = 0
         do 71 k = ia(i), ia(i+1) -1
            if (ja(k)  .lt. i) goto 71
            ko = ko+1
            ao(ko) = a(k)
            jao(ko) = ja(k)
            if (ja(k)  .eq. i) kdiag = ko
 71      continue
         if (kdiag .eq. 0 .or. kdiag .eq. kfirst) goto 72
!     exchange
         t = ao(kdiag)
         ao(kdiag) = ao(kfirst)
         ao(kfirst) = t
!     
         k = jao(kdiag)
         jao(kdiag) = jao(kfirst)
         jao(kfirst) = k
 72      iao(i) = kfirst
 7    continue
!     redefine iao(n+1)
      iao(n+1) = ko+1
      return
!----------end-of-getu ------------------------------------------------- 
!-----------------------------------------------------------------------
      end
!----------------------------------------------------------------------- 
      subroutine levels (n, jal, ial, nlev, lev, ilev, levnum)
      integer jal(*),ial(*), levnum(*), ilev(*), lev(*) 
!-----------------------------------------------------------------------
! levels gets the level structure of a lower triangular matrix 
! for level scheduling in the parallel solution of triangular systems
! strict lower matrices (e.g. unit) as well matrices with their main 
! diagonal are accepted. 
!-----------------------------------------------------------------------
! on entry:
!----------
! n        = integer. The row dimension of the matrix
! jal, ial = 
! 
! on return:
!-----------
! nlev     = integer. number of levels found
! lev      = integer array of length n containing the level
!            scheduling permutation.
! ilev     = integer array. pointer to beginning of levels in lev.
!            the numbers lev(i) to lev(i+1)-1 contain the row numbers
!            that belong to level number i, in the level scheduling
!            ordering. The equations of the same level can be solved
!            in parallel, once those of all the previous levels have
!            been solved.
! work arrays:
!-------------
! levnum   = integer array of length n (containing the level numbers
!            of each unknown on return)
!-----------------------------------------------------------------------
      do 10 i = 1, n
         levnum(i) = 0
 10   continue
!
!     compute level of each node --
!
      nlev = 0
      do 20 i = 1, n
         levi = 0
         do 15 j = ial(i), ial(i+1) - 1
            levi = max (levi, levnum(jal(j)))
 15      continue
         levi = levi+1 
         levnum(i) = levi 
         nlev = max(nlev,levi) 
 20   continue
!-------------set data structure  --------------------------------------
      do 21 j=1, nlev+1
         ilev(j) = 0
 21   continue
!------count  number   of elements in each level ----------------------- 
      do 22 j=1, n
         i = levnum(j)+1
         ilev(i) = ilev(i)+1
 22   continue
!---- set up pointer for  each  level ---------------------------------- 
      ilev(1) = 1
      do 23 j=1, nlev
         ilev(j+1) = ilev(j)+ilev(j+1)
 23   continue
!-----determine elements of each level -------------------------------- 
      do 30 j=1,n
         i = levnum(j)
         lev(ilev(i)) = j
         ilev(i) = ilev(i)+1
 30   continue
!     reset pointers backwards
      do 35 j=nlev, 1, -1
         ilev(j+1) = ilev(j) 
 35   continue
      ilev(1) = 1
      return
!----------end-of-levels------------------------------------------------ 
!-----------------------------------------------------------------------
      end
!-----------------------------------------------------------------------
      subroutine amask (nrow,ncol,a,ja,ia,jmask,imask,                    &
     &                  c,jc,ic,iw,nzmax,ierr)
!---------------------------------------------------------------------
      real*8 a(*),c(*) 
      integer ia(nrow+1),ja(*),jc(*),ic(nrow+1),jmask(*),imask(nrow+1) 
      logical iw(ncol)
!-----------------------------------------------------------------------
! This subroutine builds a sparse matrix from an input matrix by 
! extracting only elements in positions defined by the mask jmask, imask
!-----------------------------------------------------------------------
! On entry:
!---------
! nrow  = integer. row dimension of input matrix 
! ncol	= integer. Column dimension of input matrix.
!
! a,
! ja,
! ia	= matrix in Compressed Sparse Row format
!
! jmask,
! imask = matrix defining mask (pattern only) stored in compressed
!         sparse row format.
!
! nzmax = length of arrays ! and jc. see ierr.
! 
! On return:
!-----------
!
! a, ja, ia and jmask, imask are unchanged.
!
! !
! jc, 
! ic	= the output matrix in Compressed Sparse Row format.
! 
! ierr  = integer. serving as error message.!
!         ierr = 1  means normal return
!         ierr .gt. 1 means that amask stopped when processing
!         row number ierr, because there was not enough space in
!         !, jc according to the value of nzmax.
!
! work arrays:
!------------- 
! iw	= logical work array of length ncol.
!
! note: 
!------ the  algorithm is in place: !, jc, ic can be the same as 
! a, ja, ia in which cas the code will overwrite the matrix !
! on a, ja, ia
!
!-----------------------------------------------------------------------
      ierr = 0
      len = 0
      do 1 j=1, ncol
         iw(j) = .false.
 1    continue
!     unpack the mask for row ii in iw
      do 100 ii=1, nrow
!     save pointer in order to be able to do things in place
         do 2 k=imask(ii), imask(ii+1)-1
            iw(jmask(k)) = .true.
 2       continue
!     add umasked elemnts of row ii
         k1 = ia(ii)
         k2 = ia(ii+1)-1
         ic(ii) = len+1
         do 200 k=k1,k2 
            j = ja(k)
            if (iw(j)) then
               len = len+1
               if (len .gt. nzmax) then
                  ierr = ii
                  return
               endif
               jc(len) = j
               !(len) = a(k)
            endif
 200     continue	      
!     
         do 3 k=imask(ii), imask(ii+1)-1
            iw(jmask(k)) = .false.
 3       continue
 100  continue	  
      ic(nrow+1)=len+1
!
      return
!-----end-of-amask -----------------------------------------------------
!-----------------------------------------------------------------------
      end
!----------------------------------------------------------------------- 
      subroutine rperm (nrow,a,ja,ia,ao,jao,iao,perm,job)
      integer nrow,ja(*),ia(nrow+1),jao(*),iao(nrow+1),perm(nrow),job
      real*8 a(*),ao(*) 
!-----------------------------------------------------------------------
! this subroutine permutes the rows of a matrix in CSR format. 
! rperm  computes B = P A  where P is a permutation matrix.  
! the permutation P is defined through the array perm: for each j, 
! perm(j) represents the destination row number of row number j. 
! Youcef Saad -- recoded Jan 28, 1991.
!-----------------------------------------------------------------------
! on entry:
!----------
! n 	= dimension of the matrix
! a, ja, ia = input matrix in csr format
! perm 	= integer array of length nrow containing the permutation arrays
!	  for the rows: perm(i) is the destination of row i in the
!         permuted matrix. 
!         ---> a(i,j) in the original matrix becomes a(perm(i),j) 
!         in the output  matrix.
!
! job	= integer indicating the work to be done:
! 		job = 1	permute a, ja, ia into ao, jao, iao 
!                       (including the copying of real values ao and
!                       the array iao).
! 		job .ne. 1 :  ignore real values.
!                     (in which case arrays a and ao are not needed nor
!                      used).
!
!------------
! on return: 
!------------ 
! ao, jao, iao = input matrix in a, ja, ia format
! note : 
!        if (job.ne.1)  then the arrays a and ao are not used.
!----------------------------------------------------------------------!
!           Y. Saad, May  2, 1990                                      !
!----------------------------------------------------------------------!
      logical values
      values = (job .eq. 1) 
!     
!     determine pointers for output matix. 
!     
      do 50 j=1,nrow
         i = perm(j)
         iao(i+1) = ia(j+1) - ia(j)
 50   continue
!
! get pointers from lengths
!
      iao(1) = 1
      do 51 j=1,nrow
         iao(j+1)=iao(j+1)+iao(j)
 51   continue
!
! copying 
!
      do 100 ii=1,nrow
!
! old row = ii  -- new row = iperm(ii) -- ko = new pointer
!        
         ko = iao(perm(ii)) 
         do 60 k=ia(ii), ia(ii+1)-1 
            jao(ko) = ja(k) 
            if (values) ao(ko) = a(k)
            ko = ko+1
 60      continue
 100  continue
!
      return
!---------end-of-rperm ------------------------------------------------- 
!-----------------------------------------------------------------------
      end
!-----------------------------------------------------------------------
      subroutine cperm (nrow,a,ja,ia,ao,jao,iao,perm,job) 
      integer nrow,ja(*),ia(nrow+1),jao(*),iao(nrow+1),perm(*), job
      real*8 a(*), ao(*) 
!-----------------------------------------------------------------------
! this subroutine permutes the columns of a matrix a, ja, ia.
! the result is written in the output matrix  ao, jao, iao.
! cperm computes B = A P, where  P is a permutation matrix
! that maps column j into column perm(j), i.e., on return 
!      a(i,j) becomes a(i,perm(j)) in new matrix 
! Y. Saad, May 2, 1990 / modified Jan. 28, 1991. 
!-----------------------------------------------------------------------
! on entry:
!----------
! nrow 	= row dimension of the matrix
!
! a, ja, ia = input matrix in csr format. 
!
! perm	= integer array of length ncol (number of columns of A
!         containing the permutation array  the columns: 
!         a(i,j) in the original matrix becomes a(i,perm(j))
!         in the output matrix.
!
! job	= integer indicating the work to be done:
! 		job = 1	permute a, ja, ia into ao, jao, iao 
!                       (including the copying of real values ao and
!                       the array iao).
! 		job .ne. 1 :  ignore real values ao and ignore iao.
!
!------------
! on return: 
!------------ 
! ao, jao, iao = input matrix in a, ja, ia format (array ao not needed)
!
! Notes:
!------- 
! 1. if job=1 then ao, iao are not used.
! 2. This routine is in place: ja, jao can be the same. 
! 3. If the matrix is initially sorted (by increasing column number) 
!    then ao,jao,iao  may not be on return. 
! 
!----------------------------------------------------------------------!
! local parameters:
      integer k, i, nnz
!
      nnz = ia(nrow+1)-1
      do 100 k=1,nnz
         jao(k) = perm(ja(k)) 
 100  continue
!
!     done with ja array. return if no need to touch values.
!
      if (job .ne. 1) return
!
! else get new pointers -- and copy values too.
! 
      do 1 i=1, nrow+1
         iao(i) = ia(i)
 1    continue
!
      do 2 k=1, nnz
         ao(k) = a(k)
 2    continue
!
      return
!---------end-of-cperm-------------------------------------------------- 
!-----------------------------------------------------------------------
      end
!----------------------------------------------------------------------- 
      subroutine dperm (nrow,a,ja,ia,ao,jao,iao,perm,qperm,job)
      integer nrow,ja(*),ia(nrow+1),jao(*),iao(nrow+1),perm(nrow),        &
     &        qperm(*),job
      real*8 a(*),ao(*) 
!-----------------------------------------------------------------------
! This routine permutes the rows and columns of a matrix stored in CSR
! format. i.e., it computes P A Q, where P, Q are permutation matrices. 
! P maps row i into row perm(i) and Q maps column j into column qperm(j): 
!      a(i,j)    becomes   a(perm(i),qperm(j)) in new matrix
! In the particular case where Q is the transpose of P (symmetric 
! permutation of A) then qperm is not needed. 
! note that qperm should be of length ncol (number of columns) but this
! is not checked. 
!-----------------------------------------------------------------------
! Y. Saad, Sep. 21 1989 / recoded Jan. 28 1991. 
!-----------------------------------------------------------------------
! on entry: 
!---------- 
! n 	= dimension of the matrix
! a, ja, 
!    ia = input matrix in a, ja, ia format
! perm 	= integer array of length n containing the permutation arrays
!	  for the rows: perm(i) is the destination of row i in the
!         permuted matrix -- also the destination of column i in case
!         permutation is symmetric (job .le. 2) 
!
! qperm	= same thing for the columns. This should be provided only
!         if job=3 or job=4, i.e., only in the case of a nonsymmetric
!	  permutation of rows and columns. Otherwise qperm is a dummy
!
! job	= integer indicating the work to be done:
! * job = 1,2 permutation is symmetric  Ao :== P * A * transp(P)
! 		job = 1	permute a, ja, ia into ao, jao, iao 
! 		job = 2 permute matrix ignoring real values.
! * job = 3,4 permutation is non-symmetric  Ao :== P * A * Q 
! 		job = 3	permute a, ja, ia into ao, jao, iao 
! 		job = 4 permute matrix ignoring real values.
!		
! on return: 
!-----------
! ao, jao, iao = input matrix in a, ja, ia format
!
! in case job .eq. 2 or job .eq. 4, a and ao are never referred to 
! and can be dummy arguments. 
! Notes:
!------- 
!  1) algorithm is in place 
!  2) column indices may not be sorted on return even  though they may be 
!     on entry.
!----------------------------------------------------------------------!
! local variables 
      integer locjob, mod
!
!     locjob indicates whether or not real values must be copied. 
!     
      locjob = mod(job,2) 
!
! permute rows first 
! 
      call rperm (nrow,a,ja,ia,ao,jao,iao,perm,locjob)
!
! then permute columns
!
      locjob = 0
!
      if (job .le. 2) then
         call cperm (nrow,ao,jao,iao,ao,jao,iao,perm,locjob) 
      else 
         call cperm (nrow,ao,jao,iao,ao,jao,iao,qperm,locjob) 
      endif 
!     
      return
!-------end-of-dperm----------------------------------------------------
!-----------------------------------------------------------------------
      end
!-----------------------------------------------------------------------
      subroutine dperm1 (i1,i2,a,ja,ia,b,jb,ib,perm,ipos,job)
      integer i1,i2,job,ja(*),ia(*),jb(*),ib(*),perm(*)
      real*8 a(*),b(*)
!----------------------------------------------------------------------- 
!     general submatrix extraction routine.
!----------------------------------------------------------------------- 
!     extracts rows perm(i1), perm(i1+1), ..., perm(i2) (in this order) 
!     from a matrix (doing nothing in the column indices.) The resulting 
!     submatrix is constructed in b, jb, ib. A pointer ipos to the
!     beginning of arrays b,jb,is also allowed (i.e., nonzero elements
!     are accumulated starting in position ipos of b, jb). 
!-----------------------------------------------------------------------
! Y. Saad,Sep. 21 1989 / recoded Jan. 28 1991 / modified for PSPARSLIB 
! Sept. 1997.. 
!-----------------------------------------------------------------------
! on entry: 
!---------- 
! n 	= dimension of the matrix
! a,ja,
!   ia  = input matrix in CSR format
! perm 	= integer array of length n containing the indices of the rows
!         to be extracted. 
!
! job   = job indicator. if (job .ne.1) values are not copied (i.e.,
!         only pattern is copied).
!
! on return: 
!-----------
! b,ja,
! ib   = matrix in csr format. b(ipos:ipos+nnz-1),jb(ipos:ipos+nnz-1) 
!     contain the value and column indices respectively of the nnz
!     nonzero elements of the permuted matrix. thus ib(1)=ipos.
!
! Notes:
!------- 
!  algorithm is NOT in place 
!----------------------------------------------------------------------- 
! local variables
!
      integer ko,irow,k 
      logical values
!-----------------------------------------------------------------------
      values = (job .eq. 1) 
      ko = ipos 
      ib(1) = ko
      do 900 i=i1,i2
         irow = perm(i) 
         do 800 k=ia(irow),ia(irow+1)-1
            if (values) b(ko) = a(k)
            jb(ko) = ja(k)
            ko=ko+1
 800     continue
         ib(i-i1+2) = ko
 900  continue
      return
!--------end-of-dperm1--------------------------------------------------
!-----------------------------------------------------------------------
      end
!-----------------------------------------------------------------------
      subroutine dperm2 (i1,i2,a,ja,ia,b,jb,ib,cperm,rperm,istart,        &
     &        ipos,job)
      integer i1,i2,job,istart,ja(*),ia(*),jb(*),ib(*),                   &
     & cperm(*),rperm(*) 
      real*8 a(*),b(*)
!----------------------------------------------------------------------- 
!     general submatrix permutation/ extraction routine.
!----------------------------------------------------------------------- 
!     extracts rows rperm(i1), rperm(i1+1), ..., rperm(i2) and does an
!     associated column permutation (using array cperm). The resulting
!     submatrix is constructed in b, jb, ib. For added flexibility, the
!     extracted elements are put in sequence starting from row 'istart' 
!     of B. In addition a pointer ipos to the beginning of arrays b,jb,
!     is also allowed (i.e., nonzero elements are accumulated starting in
!     position ipos of b, jb). In most applications istart and ipos are 
!     equal to one. However, the generality adds substantial flexiblity.
!     EXPLE: (1) to permute msr to msr (excluding diagonals) 
!     call dperm2 (1,n,a,ja,ja,b,jb,jb,rperm,rperm,1,n+2) 
!            (2) To extract rows 1 to 10: define rperm and cperm to be
!     identity permutations (rperm(i)=i, i=1,n) and then
!            call dperm2 (1,10,a,ja,ia,b,jb,ib,rperm,rperm,1,1) 
!            (3) to achieve a symmetric permutation as defined by perm: 
!            call dperm2 (1,10,a,ja,ia,b,jb,ib,perm,perm,1,1) 
!            (4) to get a symmetric permutation of A and append the
!            resulting data structure to A's data structure (useful!) 
!            call dperm2 (1,10,a,ja,ia,a,ja,ia(n+1),perm,perm,1,ia(n+1))
!-----------------------------------------------------------------------
! Y. Saad,Sep. 21 1989 / recoded Jan. 28 1991. 
!-----------------------------------------------------------------------
! on entry: 
!---------- 
! n 	= dimension of the matrix
! i1,i2 = extract rows rperm(i1) to rperm(i2) of A, with i1<i2.
!
! a,ja,
!   ia  = input matrix in CSR format
! cperm = integer array of length n containing the permutation arrays
!	  for the columns: cperm(i) is the destination of column j, 
!         i.e., any column index ja(k) is transformed into cperm(ja(k)) 
!
! rperm	=  permutation array for the rows. rperm(i) = origin (in A) of
!          row i in B. This is the reverse permutation relative to the
!          ones used in routines cperm, dperm,.... 
!          rows rperm(i1), rperm(i1)+1, ... rperm(i2) are 
!          extracted from A and stacked into B, starting in row istart
!          of B. 
! istart= starting row for B where extracted matrix is to be added.
!         this is also only a pointer of the be beginning address for
!         ib , on return. 
! ipos  = beginning position in arrays b and jb where to start copying 
!         elements. Thus, ib(istart) = ipos. 
!
! job   = job indicator. if (job .ne.1) values are not copied (i.e.,
!         only pattern is copied).
!
! on return: 
!-----------
! b,ja,
! ib   = matrix in csr format. positions 1,2,...,istart-1 of ib 
!     are not touched. b(ipos:ipos+nnz-1),jb(ipos:ipos+nnz-1) 
!     contain the value and column indices respectively of the nnz
!     nonzero elements of the permuted matrix. thus ib(istart)=ipos.
!
! Notes:
!------- 
!  1) algorithm is NOT in place 
!  2) column indices may not be sorted on return even  though they 
!     may be on entry.
!----------------------------------------------------------------------- 
! local variables
!
      integer ko,irow,k 
      logical values
!-----------------------------------------------------------------------
      values = (job .eq. 1) 
      ko = ipos 
      ib(istart) = ko
      do 900 i=i1,i2
         irow = rperm(i) 
         do 800 k=ia(irow),ia(irow+1)-1
            if (values) b(ko) = a(k)
            jb(ko) = cperm(ja(k))
            ko=ko+1
 800     continue
         ib(istart+i-i1+1) = ko
 900  continue
      return
!--------end-of-dperm2--------------------------------------------------
!-----------------------------------------------------------------------
      end 
!-----------------------------------------------------------------------
      subroutine dmperm (nrow,a,ja,ao,jao,perm,job)
      integer nrow,ja(*),jao(*),perm(nrow),job
      real*8 a(*),ao(*) 
!-----------------------------------------------------------------------
! This routine performs a symmetric permutation of the rows and 
! columns of a matrix stored in MSR format. i.e., it computes 
! B = P A transp(P), where P, is  a permutation matrix. 
! P maps row i into row perm(i) and column j into column perm(j): 
!      a(i,j)    becomes   a(perm(i),perm(j)) in new matrix
! (i.e.  ao(perm(i),perm(j)) = a(i,j) ) 
! calls dperm. 
!-----------------------------------------------------------------------
! Y. Saad, Nov 15, 1991. 
!-----------------------------------------------------------------------
! on entry: 
!---------- 
! n 	= dimension of the matrix
! a, ja = input matrix in MSR format. 
! perm 	= integer array of length n containing the permutation arrays
!	  for the rows: perm(i) is the destination of row i in the
!         permuted matrix -- also the destination of column i in case
!         permutation is symmetric (job .le. 2) 
!
! job	= integer indicating the work to be done:
! 		job = 1	permute a, ja, ia into ao, jao, iao 
! 		job = 2 permute matrix ignoring real values.
!		
! on return: 
!-----------
! ao, jao = output matrix in MSR. 
!
! in case job .eq. 2 a and ao are never referred to and can be dummy 
! arguments. 
!
! Notes:
!------- 
!  1) algorithm is NOT in place 
!  2) column indices may not be sorted on return even  though they may be 
!     on entry.
!----------------------------------------------------------------------!
!     local variables
!     
      integer n1, n2
      n1 = nrow+1
      n2 = n1+1
!      
      call dperm (nrow,a,ja,ja,ao(n2),jao(n2),jao,perm,perm,job) 
!     
      jao(1) = n2
      do 101 j=1, nrow 
         ao(perm(j)) = a(j) 
         jao(j+1) = jao(j+1)+n1
 101  continue
!
! done
!     
      return
!-----------------------------------------------------------------------
!--------end-of-dmperm--------------------------------------------------
      end
!----------------------------------------------------------------------- 
      subroutine dvperm (n, x, perm) 
      integer n, perm(n) 
      real*8 x(n)
!-----------------------------------------------------------------------
! this subroutine performs an in-place permutation of a real vector x 
! according to the permutation array perm(*), i.e., on return, 
! the vector x satisfies,
!
!	x(perm(j)) :== x(j), j=1,2,.., n
!
!-----------------------------------------------------------------------
! on entry:
!---------
! n 	= length of vector x.
! perm 	= integer array of length n containing the permutation  array.
! x	= input vector
!
! on return:
!---------- 
! x	= vector x permuted according to x(perm(*)) :=  x(*)
!
!----------------------------------------------------------------------!
!           Y. Saad, Sep. 21 1989                                      !
!----------------------------------------------------------------------!
! local variables 
      real*8 tmp, tmp1
!
      init      = 1
      tmp	= x(init)	
      ii        = perm(init)
      perm(init)= -perm(init)
      k         = 0
!     
! loop
! 
 6    k = k+1
!
! save the chased element --
! 
      tmp1	  = x(ii) 
      x(ii)     = tmp
      next	  = perm(ii) 
      if (next .lt. 0 ) goto 65
!     
! test for end 
!
      if (k .gt. n) goto 101
      tmp       = tmp1
      perm(ii)  = - perm(ii)
      ii        = next 
!
! end loop 
!
      goto 6
!
! reinitilaize cycle --
!
 65   init      = init+1
      if (init .gt. n) goto 101
      if (perm(init) .lt. 0) goto 65
      tmp	= x(init)
      ii	= perm(init)
      perm(init)=-perm(init)
      goto 6
!     
 101  continue
      do 200 j=1, n
         perm(j) = -perm(j)
 200  continue 
!     
      return
!-------------------end-of-dvperm--------------------------------------- 
!-----------------------------------------------------------------------
      end
!-----------------------------------------------------------------------
      subroutine ivperm (n, ix, perm) 
      integer n, perm(n), ix(n)
!-----------------------------------------------------------------------
! this subroutine performs an in-place permutation of an integer vector 
! ix according to the permutation array perm(*), i.e., on return, 
! the vector x satisfies,
!
!	ix(perm(j)) :== ix(j), j=1,2,.., n
!
!-----------------------------------------------------------------------
! on entry:
!---------
! n 	= length of vector x.
! perm 	= integer array of length n containing the permutation  array.
! ix	= input vector
!
! on return:
!---------- 
! ix	= vector x permuted according to ix(perm(*)) :=  ix(*)
!
!----------------------------------------------------------------------!
!           Y. Saad, Sep. 21 1989                                      !
!----------------------------------------------------------------------!
! local variables
      integer tmp, tmp1
!
      init      = 1
      tmp	= ix(init)	
      ii        = perm(init)
      perm(init)= -perm(init)
      k         = 0
!     
! loop
! 
 6    k = k+1
!
! save the chased element --
! 
      tmp1	  = ix(ii) 
      ix(ii)     = tmp
      next	  = perm(ii) 
      if (next .lt. 0 ) goto 65
!     
! test for end 
!
      if (k .gt. n) goto 101
      tmp       = tmp1
      perm(ii)  = - perm(ii)
      ii        = next 
!
! end loop 
!
      goto 6
!
! reinitilaize cycle --
!
 65   init      = init+1
      if (init .gt. n) goto 101
      if (perm(init) .lt. 0) goto 65
      tmp	= ix(init)
      ii	= perm(init)
      perm(init)=-perm(init)
      goto 6
!     
 101  continue
      do 200 j=1, n
         perm(j) = -perm(j)
 200  continue 
!     
      return
!-------------------end-of-ivperm--------------------------------------- 
!-----------------------------------------------------------------------
      end
!-----------------------------------------------------------------------  
      subroutine retmx (n,a,ja,ia,dd)
      real*8 a(*),dd(*)
      integer n,ia(*),ja(*)
!-----------------------------------------------------------------------
! returns in dd(*) the max absolute value of elements in row *.
! used for scaling purposes. superseded by rnrms  .
!
! on entry:
! n	= dimension of A
! a,ja,ia
!	= matrix stored in compressed sparse row format
! dd	= real*8 array of length n. On output,entry dd(i) contains
!	  the element of row i that has the largest absolute value.
!	  Moreover the sign of dd is modified such that it is the
!	  same as that of the diagonal element in row i.
!----------------------------------------------------------------------!
!           Y. Saad, Sep. 21 1989                                      !
!----------------------------------------------------------------------!
! local variables
      integer k2, i, k1, k
      real*8 t, t1, t2
!
! initialize 
!
      k2 = 1
      do 11 i=1,n
         k1 = k2
         k2 = ia(i+1) - 1
         t = 0.0d0
         do 101  k=k1,k2
            t1 = abs(a(k))
            if (t1 .gt. t) t = t1
            if (ja(k) .eq. i) then 
               if (a(k) .ge. 0.0) then 
                  t2 = a(k) 
               else 
                  t2 = - a(k)
               endif
            endif
 101     continue		
         dd(i) =  t2*t
!     we do not invert diag
 11   continue
      return
!---------end of retmx -------------------------------------------------
!-----------------------------------------------------------------------
      end
!----------------------------------------------------------------------- 
      subroutine diapos  (n,ja,ia,idiag) 
      integer ia(n+1), ja(*), idiag(n) 
!-----------------------------------------------------------------------
! this subroutine returns the positions of the diagonal elements of a
! sparse matrix a, ja, ia, in the array idiag.
!-----------------------------------------------------------------------
! on entry:
!---------- 
!
! n	= integer. row dimension of the matrix a.
! a,ja,
!    ia = matrix stored compressed sparse row format. a array skipped.
!
! on return:
!-----------
! idiag  = integer array of length n. The i-th entry of idiag 
!          points to the diagonal element a(i,i) in the arrays
!          a, ja. (i.e., a(idiag(i)) = element A(i,i) of matrix A)
!          if no diagonal element is found the entry is set to 0.
!----------------------------------------------------------------------!
!           Y. Saad, March, 1990
!----------------------------------------------------------------------!
      do 1 i=1, n 
         idiag(i) = 0
 1    continue
!     
!     sweep through data structure. 
!     
      do  6 i=1,n
         do 51 k= ia(i),ia(i+1) -1
            if (ja(k) .eq. i) idiag(i) = k
 51      continue
 6    continue
!----------- -end-of-diapos---------------------------------------------
!-----------------------------------------------------------------------
      return
      end
!-----------------------------------------------------------------------
      subroutine dscaldg (n,a,ja,ia,diag,job)
      real*8 a(*), diag(*),t
      integer ia(*),ja(*)
!----------------------------------------------------------------------- 
! scales rows by diag where diag is either given (job=0)
! or to be computed:
!  job = 1 ,scale row i by  by  +/- max |a(i,j) | and put inverse of 
!       scaling factor in diag(i),where +/- is the sign of a(i,i).
!  job = 2 scale by 2-norm of each row..
! if diag(i) = 0,then diag(i) is replaced by one
! (no scaling)..
!----------------------------------------------------------------------!
!           Y. Saad, Sep. 21 1989                                      !
!----------------------------------------------------------------------!
      goto (12,11,10) job+1
 10   do 110 j=1,n
         k1= ia(j)
         k2 = ia(j+1)-1
         t = 0.0d0
         do 111 k = k1,k2
 111        t = t+a(k)*a(k)
 110        diag(j) = sqrt(t)
            goto 12
 11   continue
      call retmx (n,a,ja,ia,diag)
!------
 12   do 1 j=1,n
         if (diag(j) .ne. 0.0d0) then 
            diag(j) = 1.0d0/diag(j)
         else 
            diag(j) = 1.0d0
         endif
 1    continue
      do 2 i=1,n
         t = diag(i)
         do 21 k=ia(i),ia(i+1) -1
            a(k) = a(k)*t
 21      continue
 2    continue
      return 
!--------end of dscaldg -----------------------------------------------
!-----------------------------------------------------------------------
      end
!-----------------------------------------------------------------------
      subroutine extbdg (n,a,ja,ia,bdiag,nblk,ao,jao,iao)
      implicit real*8 (a-h,o-z)
      real*8 bdiag(*),a(*),ao(*)
      integer ia(*),ja(*),jao(*),iao(*) 
!-----------------------------------------------------------------------
! this subroutine extracts the main diagonal blocks of a 
! matrix stored in compressed sparse row format and puts the result
! into the array bdiag and the remainder in ao,jao,iao.
!-----------------------------------------------------------------------
! on entry:
!----------
! n	= integer. The row dimension of the matrix a.
! a,
! ja,
! ia    = matrix stored in csr format
! nblk  = dimension of each diagonal block. The diagonal blocks are
!         stored in compressed format rowwise,i.e.,we store in 
!	  succession the i nonzeros of the i-th row after those of
!	  row number i-1..
!
! on return:
!----------
! bdiag = real*8 array of size (n x nblk) containing the diagonal
!	  blocks of A on return
! ao,
! jao,
! iao   = remainder of the matrix stored in csr format.
!----------------------------------------------------------------------!
!           Y. Saad, Sep. 21 1989                                      !
!----------------------------------------------------------------------!
      m = 1 + (n-1)/nblk
! this version is sequential -- there is a more parallel version
! that goes through the structure twice ....
      ltr =  ((nblk-1)*nblk)/2 
      l = m * ltr
      do 1 i=1,l
         bdiag(i) = 0.0d0
 1    continue
      ko = 0
      kb = 1
      iao(1) = 1
!-------------------------
      do 11 jj = 1,m
         j1 = (jj-1)*nblk+1
         j2 =  min0 (n,j1+nblk-1)
         do 12 j=j1,j2
            do 13 i=ia(j),ia(j+1) -1
               k = ja(i)
               if (k .lt. j1) then
                  ko = ko+1
                  ao(ko) = a(i)
                  jao(ko) = k
               else if (k .lt. j) then
!     kb = (jj-1)*ltr+((j-j1)*(j-j1-1))/2+k-j1+1
!     bdiag(kb) = a(i)
                  bdiag(kb+k-j1) = a(i)
               endif
 13         continue 	
            kb = kb + j-j1
            iao(j+1) = ko+1
 12      continue
 11   continue
      return
!---------end-of-extbdg------------------------------------------------- 
!----------------------------------------------------------------------- 
      end
!----------------------------------------------------------------------- 
      subroutine getbwd(n,a,ja,ia,ml,mu)
!-----------------------------------------------------------------------
! gets the bandwidth of lower part and upper part of A.
! does not assume that A is sorted.
!-----------------------------------------------------------------------
! on entry:
!----------
! n	= integer = the row dimension of the matrix
! a, ja,
!    ia = matrix in compressed sparse row format.
! 
! on return:
!----------- 
! ml	= integer. The bandwidth of the strict lower part of A
! mu	= integer. The bandwidth of the strict upper part of A 
!
! Notes:
! ===== ml and mu are allowed to be negative or return. This may be 
!       useful since it will tell us whether a band is confined 
!       in the strict  upper/lower triangular part. 
!       indeed the definitions of ml and mu are
!
!       ml = max ( (i-j)  s.t. a(i,j) .ne. 0  )
!       mu = max ( (j-i)  s.t. a(i,j) .ne. 0  )
!----------------------------------------------------------------------!
! Y. Saad, Sep. 21 1989                                                !
!----------------------------------------------------------------------!
      real*8 a(*) 
      integer ja(*),ia(n+1),ml,mu,ldist,i,k 
      ml = - n
      mu = - n
      do 3 i=1,n
         do 31 k=ia(i),ia(i+1)-1 
            ldist = i-ja(k)
            ml = max(ml,ldist)
            mu = max(mu,-ldist)
 31      continue
 3    continue
      return
!---------------end-of-getbwd ------------------------------------------ 
!----------------------------------------------------------------------- 
      end
!----------------------------------------------------------------------- 
      subroutine blkfnd (nrow,ja,ia,nblk)
!-----------------------------------------------------------------------
! This routine attemptps to determine whether or not  the input
! matrix has a block structure and finds the blocks size
! if it does. A block matrix is one which is
! comprised of small square dense blocks. If there are zero
! elements within the square blocks and the original data structure
! takes these zeros into account then blkchk may fail to find the
! correct block size. 
!----------------------------------------------------------------------- 
! on entry
!---------
! nrow	= integer equal to the row dimension of the matrix.  
! ja    = integer array containing the column indices of the entries 
!         nonzero entries of the matrix stored by row.
! ia    = integer array of length nrow + 1 containing the pointers 
!         beginning of each row in array ja.
!		
! nblk  = integer containing the assumed value of nblk if job = 0
!         
! on return
!---------- 
! nblk  = integer containing the value found for nblk when job = 1.
!         if imsg .ne. 0 this value is meaningless however.
!
!----------------------------------------------------------------------!
!           Y. Saad, Sep. 21 1989                                      !
!----------------------------------------------------------------------!
      integer ia(nrow+1),ja(*) 
!----------------------------------------------------------------------- 
! first part of code will find candidate block sizes.
! criterion used here is a simple one: scan rows and  determine groups 
! of rows that have the same length and such that the first column 
! number and the last column number are identical. 
!----------------------------------------------------------------------- 
      minlen = ia(2)-ia(1)
      irow   = 1
      do 1 i=2,nrow
         len = ia(i+1)-ia(i)
         if (len .lt. minlen) then
            minlen = len 
            irow = i
         endif
 1    continue
!     
!     ---- candidates are all dividers of minlen
!
      nblk = 1
      if (minlen .le. 1) return
!     
      do 99 iblk = minlen, 1, -1
         if (mod(minlen,iblk) .ne. 0) goto 99
         len = ia(2) - ia(1)
         len0 = len
         jfirst = ja(1) 
         jlast = ja(ia(2)-1)
         do 10 jrow = irow+1,irow+nblk-1
            i1 = ia(jrow)
            i2 = ia(jrow+1)-1
            len = i2+1-i1
            jf = ja(i1)
            jl = ja(i2) 
            if (len .ne. len0 .or. jf .ne. jfirst .or.                    &
     &           jl .ne. jlast) goto 99
 10      continue
!     
!     check for this candidate ----
!     
         call blkchk (nrow,ja,ia,iblk,imsg)	       
         if (imsg .eq. 0) then 
!     
!     block size found
!     
            nblk = iblk
            return
         endif 
 99   continue	       
!--------end-of-blkfnd ------------------------------------------------- 
!-----------------------------------------------------------------------
      end
!-----------------------------------------------------------------------
      subroutine blkchk (nrow,ja,ia,nblk,imsg)
!----------------------------------------------------------------------- 
! This routine checks whether the input matrix is a block
! matrix with block size of nblk. A block matrix is one which is
! comprised of small square dense blocks. If there are zero
! elements within the square blocks and the data structure
! takes them into account then blkchk may fail to find the
! correct block size. 
!----------------------------------------------------------------------- 
! on entry
!---------
! nrow	= integer equal to the row dimension of the matrix.  
! ja    = integer array containing the column indices of the entries 
!         nonzero entries of the matrix stored by row.
! ia    = integer array of length nrow + 1 containing the pointers 
!         beginning of each row in array ja.
!
! nblk  = integer containing the value of nblk to be checked. 
!         
! on return
!---------- 
!
! imsg  = integer containing a message  with the following meaning.
!          imsg = 0 means that the output value of nblk is a correct
!                   block size. nblk .lt. 0 means nblk not correct 
!                   block size.
!          imsg = -1 : nblk does not divide nrow 
!          imsg = -2 : a starting element in a row is at wrong position
!             (j .ne. mult*nblk +1 )
!          imsg = -3 : nblk does divide a row length - 
!          imsg = -4 : an element is isolated outside a block or
!             two rows in same group have different lengths
!----------------------------------------------------------------------!
!           Y. Saad, Sep. 21 1989                                      !
!----------------------------------------------------------------------!
      integer ia(nrow+1),ja(*) 
!---------------------------------------------------------------------- 
! first part of code will find candidate block sizes.
! this is not guaranteed to work . so a check is done at the end
! the criterion used here is a simple one:
! scan rows and determine groups of rows that have the same length
! and such that the first column number and the last column number 
! are identical. 
!---------------------------------------------------------------------- 
      imsg = 0
      if (nblk .le. 1) return
      nr = nrow/nblk
      if (nr*nblk .ne. nrow) goto 101
!--   main loop --------------------------------------------------------- 
      irow = 1
      do 20 ii=1, nr
!     i1= starting position for group of nblk rows in original matrix
         i1 = ia(irow)
         j2 = i1
!     lena = length of each row in that group  in the original matrix
         lena = ia(irow+1)-i1
!     len = length of each block-row in that group in the output matrix
         len = lena/nblk
         if (len* nblk .ne. lena) goto 103
!     
!     for each row
!     
         do 6 i = 1, nblk
            irow = irow + 1
            if (ia(irow)-ia(irow-1) .ne. lena ) goto 104
!     
!     for each block
!     
            do 7 k=0, len-1
               jstart = ja(i1+nblk*k)-1
               if ( (jstart/nblk)*nblk .ne. jstart) goto 102
!     
!     for each column
!     
               do 5 j=1, nblk
                  if (jstart+j .ne. ja(j2) )  goto 104
                  j2 = j2+1 
 5             continue
 7          continue
 6       continue
 20   continue
!     went through all loops successfully:
      return
 101  imsg = -1
      return
 102  imsg = -2
      return
 103  imsg = -3
      return
 104  imsg = -4
!----------------end of chkblk ----------------------------------------- 
!-----------------------------------------------------------------------
      return
      end
!-----------------------------------------------------------------------
      subroutine infdia (n,ja,ia,ind,idiag) 
      integer ia(*), ind(*), ja(*)
!-----------------------------------------------------------------------
!     obtains information on the diagonals of A. 
!----------------------------------------------------------------------- 
! this subroutine finds the lengths of each of the 2*n-1 diagonals of A
! it also outputs the number of nonzero diagonals found. 
!----------------------------------------------------------------------- 
! on entry:
!---------- 
! n	= dimension of the matrix a.
!
! a,    ..... not needed here.
! ja, 			
! ia    = matrix stored in csr format
!
! on return:
!----------- 
!
! idiag = integer. number of nonzero diagonals found. 
! 
! ind   = integer array of length at least 2*n-1. The k-th entry in
!         ind contains the number of nonzero elements in the diagonal
!         number k, the numbering beeing from the lowermost diagonal
!         (bottom-left). In other words ind(k) = length of diagonal
!         whose offset wrt the main diagonal is = - n + k.
!----------------------------------------------------------------------!
!           Y. Saad, Sep. 21 1989                                      !
!----------------------------------------------------------------------!
      n2= n+n-1
      do 1 i=1,n2
         ind(i) = 0
 1    continue
      do 3 i=1, n
         do 2 k=ia(i),ia(i+1)-1
            j = ja(k)
            ind(n+j-i) = ind(n+j-i) +1
 2       continue 
 3    continue
!     count the nonzero ones.
      idiag = 0 
      do 41 k=1, n2
         if (ind(k) .ne. 0) idiag = idiag+1
 41   continue
      return
! done
!------end-of-infdia ---------------------------------------------------
!-----------------------------------------------------------------------
      end
!----------------------------------------------------------------------- 
      subroutine amubdg (nrow,ncol,ncolb,ja,ia,jb,ib,ndegr,nnz,iw) 
      integer ja(*),jb(*),ia(nrow+1),ib(ncol+1),ndegr(nrow),iw(ncolb) 
!-----------------------------------------------------------------------
! gets the number of nonzero elements in each row of A*B and the total 
! number of nonzero elements in A*B. 
!-----------------------------------------------------------------------
! on entry:
! -------- 
!
! nrow  = integer.  row dimension of matrix A
! ncol  = integer.  column dimension of matrix A = row dimension of 
!                   matrix B.
! ncolb = integer. the colum dimension of the matrix B.
!
! ja, ia= row structure of input matrix A: ja = column indices of
!         the nonzero elements of A stored by rows.
!         ia = pointer to beginning of each row  in ja.
! 
! jb, ib= row structure of input matrix B: jb = column indices of
!         the nonzero elements of A stored by rows.
!         ib = pointer to beginning of each row  in jb.
! 
! on return:
! ---------
! ndegr	= integer array of length nrow containing the degrees (i.e., 
!         the number of nonzeros in  each row of the matrix A * B 
!				
! nnz   = total number of nonzero elements found in A * B
!
! work arrays:
!-------------
! iw	= integer work array of length ncolb. 
!-----------------------------------------------------------------------
      do 1 k=1, ncolb 
         iw(k) = 0 
 1    continue
      
      do 2 k=1, nrow
         ndegr(k) = 0 
 2    continue
!     
!     method used: Transp(A) * A = sum [over i=1, nrow]  a(i)^T a(i)
!     where a(i) = i-th row of  A. We must be careful not to add  the
!     elements already accounted for.
!     
!     
      do 7 ii=1,nrow 
!     
!     for each row of A
!     
         ldg = 0 
!     
!    end-of-linked list
!     
         last = -1 
         do 6 j = ia(ii),ia(ii+1)-1 
!     
!     row number to be added:
!     
            jr = ja(j) 
            do 5 k=ib(jr),ib(jr+1)-1
               jc = jb(k) 
               if (iw(jc) .eq. 0) then 
!     
!     add one element to the linked list 
!     
                  ldg = ldg + 1
                  iw(jc) = last 
                  last = jc
               endif
 5          continue
 6       continue
         ndegr(ii) = ldg
!     
!     reset iw to zero
!     
         do 61 k=1,ldg 
            j = iw(last) 
            iw(last) = 0
            last = j
 61      continue
!-----------------------------------------------------------------------
 7    continue
!     
      nnz = 0
      do 8 ii=1, nrow 
         nnz = nnz+ndegr(ii) 
 8    continue
!     
      return
!---------------end-of-amubdg ------------------------------------------
!-----------------------------------------------------------------------
      end
!----------------------------------------------------------------------- 
      subroutine aplbdg (nrow,ncol,ja,ia,jb,ib,ndegr,nnz,iw) 
      integer ja(*),jb(*),ia(nrow+1),ib(nrow+1),iw(ncol),ndegr(nrow) 
!-----------------------------------------------------------------------
! gets the number of nonzero elements in each row of A+B and the total 
! number of nonzero elements in A+B. 
!-----------------------------------------------------------------------
! on entry:
! ---------
! nrow	= integer. The row dimension of A and B
! ncol  = integer. The column dimension of A and B.
!
! a,
! ja,
! ia   = Matrix A in compressed sparse row format.
! 
! b, 
! jb, 
! ib	=  Matrix B in compressed sparse row format.
!
! on return:
!----------
! ndegr	= integer array of length nrow containing the degrees (i.e., 
!         the number of nonzeros in  each row of the matrix A + B.
!				
! nnz   = total number of nonzero elements found in A * B
!
! work arrays:
!------------
! iw	= integer work array of length equal to ncol. 
!
!-----------------------------------------------------------------------
      do 1 k=1, ncol 
         iw(k) = 0 
 1    continue
!
      do 2 k=1, nrow
         ndegr(k) = 0 
 2    continue
!
      do 7 ii=1,nrow 
         ldg = 0 
!     
!    end-of-linked list
!     
         last = -1 
!     
!     row of A
!     
         do 5 j = ia(ii),ia(ii+1)-1 
            jr = ja(j) 
!     
!     add element to the linked list 
!     
            ldg = ldg + 1
            iw(jr) = last 
            last = jr
 5       continue
!     
!     row of B
!     
         do 6 j=ib(ii),ib(ii+1)-1
            jc = jb(j)
            if (iw(jc) .eq. 0) then 
!     
!     add one element to the linked list 
!     
               ldg = ldg + 1
               iw(jc) = last 
               last = jc
            endif
 6       continue
!     done with row ii. 
         ndegr(ii) = ldg
!     
!     reset iw to zero
!     
         do 61 k=1,ldg 
            j = iw(last) 
            iw(last) = 0
            last = j
 61      continue
!-----------------------------------------------------------------------
 7    continue
!     
      nnz = 0
      do 8 ii=1, nrow 
         nnz = nnz+ndegr(ii) 
 8    continue
      return
!----------------end-of-aplbdg -----------------------------------------
!-----------------------------------------------------------------------
      end
!-----------------------------------------------------------------------
       subroutine rnrms(nrow, nrm, a, ja, ia, diag) 
      real*8 a(*), diag(nrow), scal 
      integer ja(*), ia(nrow+1) 
!-----------------------------------------------------------------------
! gets the norms of each row of A. (choice of three norms)
!-----------------------------------------------------------------------
! on entry:
! ---------
! nrow	= integer. The row dimension of A
!
! nrm   = integer. norm indicator. nrm = 1, means 1-norm, nrm =2
!                  means the 2-nrm, nrm = 0 means max norm
!
! a,
! ja,
! ia   = Matrix A in compressed sparse row format.
! 
! on return:
!----------
!
! diag = real vector of length nrow containing the norms
!
!-----------------------------------------------------------------
      do 1 ii=1,nrow
!
!     compute the norm if each element.
!     
         scal = 0.0d0
         k1 = ia(ii)
         k2 = ia(ii+1)-1
         if (nrm .eq. 0) then
            do 2 k=k1, k2
               scal = max(scal,abs(a(k) ) ) 
 2          continue
         elseif (nrm .eq. 1) then
            do 3 k=k1, k2
               scal = scal + abs(a(k) ) 
 3          continue
         else
            do 4 k=k1, k2
               scal = scal+a(k)**2
 4          continue
         endif 
         if (nrm .eq. 2) scal = sqrt(scal) 
         diag(ii) = scal
 1    continue
      return
!-----------------------------------------------------------------------
!-------------end-of-rnrms----------------------------------------------
      end 
!----------------------------------------------------------------------- 
      subroutine cnrms   (nrow, nrm, a, ja, ia, diag) 
      real*8 a(*), diag(nrow) 
      integer ja(*), ia(nrow+1) 
!-----------------------------------------------------------------------
! gets the norms of each column of A. (choice of three norms)
!-----------------------------------------------------------------------
! on entry:
! ---------
! nrow	= integer. The row dimension of A
!
! nrm   = integer. norm indicator. nrm = 1, means 1-norm, nrm =2
!                  means the 2-nrm, nrm = 0 means max norm
!
! a,
! ja,
! ia   = Matrix A in compressed sparse row format.
! 
! on return:
!----------
!
! diag = real vector of length nrow containing the norms
!
!-----------------------------------------------------------------
      do 10 k=1, nrow 
         diag(k) = 0.0d0
 10   continue
      do 1 ii=1,nrow
         k1 = ia(ii)
         k2 = ia(ii+1)-1
         do 2 k=k1, k2
            j = ja(k) 
!     update the norm of each column
            if (nrm .eq. 0) then
               diag(j) = max(diag(j),abs(a(k) ) ) 
            elseif (nrm .eq. 1) then
               diag(j) = diag(j) + abs(a(k) ) 
            else
               diag(j) = diag(j)+a(k)**2
            endif 
 2       continue
 1    continue
      if (nrm .ne. 2) return
      do 3 k=1, nrow
         diag(k) = sqrt(diag(k))
 3    continue
      return
!-----------------------------------------------------------------------
!------------end-of-cnrms-----------------------------------------------
      end 
!----------------------------------------------------------------------- 
      subroutine roscal(nrow,job,nrm,a,ja,ia,diag,b,jb,ib,ierr) 
      real*8 a(*), b(*), diag(nrow) 
      integer nrow,job,nrm,ja(*),jb(*),ia(nrow+1),ib(nrow+1),ierr 
!-----------------------------------------------------------------------
! scales the rows of A such that their norms are one on return
! 3 choices of norms: 1-norm, 2-norm, max-norm.
!-----------------------------------------------------------------------
! on entry:
! ---------
! nrow	= integer. The row dimension of A
!
! job   = integer. job indicator. Job=0 means get array b only
!         job = 1 means get b, and the integer arrays ib, jb.
!
! nrm   = integer. norm indicator. nrm = 1, means 1-norm, nrm =2
!                  means the 2-nrm, nrm = 0 means max norm
!
! a,
! ja,
! ia   = Matrix A in compressed sparse row format.
! 
! on return:
!----------
!
! diag = diagonal matrix stored as a vector containing the matrix
!        by which the rows have been scaled, i.e., on return 
!        we have B = Diag*A.
!
! b, 
! jb, 
! ib	= resulting matrix B in compressed sparse row sparse format.
!	    
! ierr  = error message. ierr=0     : Normal return 
!                        ierr=i > 0 : Row number i is a zero row.
! Notes:
!-------
! 1)        The column dimension of A is not needed. 
! 2)        algorithm in place (B can take the place of A).
!-----------------------------------------------------------------
      call rnrms (nrow,nrm,a,ja,ia,diag)
      ierr = 0
      do 1 j=1, nrow
         if (diag(j) .eq. 0.0d0) then
            ierr = j 
            return
         else
            diag(j) = 1.0d0/diag(j)
         endif
 1    continue
      call diamua(nrow,job,a,ja,ia,diag,b,jb,ib)
      return
!-------end-of-roscal---------------------------------------------------
!-----------------------------------------------------------------------
      end
!-----------------------------------------------------------------------
      subroutine coscal(nrow,job,nrm,a,ja,ia,diag,b,jb,ib,ierr) 
!----------------------------------------------------------------------- 
      real*8 a(*),b(*),diag(nrow) 
      integer nrow,job,ja(*),jb(*),ia(nrow+1),ib(nrow+1),ierr 
!-----------------------------------------------------------------------
! scales the columns of A such that their norms are one on return
! result matrix written on b, or overwritten on A.
! 3 choices of norms: 1-norm, 2-norm, max-norm. in place.
!-----------------------------------------------------------------------
! on entry:
! ---------
! nrow	= integer. The row dimension of A
!
! job   = integer. job indicator. Job=0 means get array b only
!         job = 1 means get b, and the integer arrays ib, jb.
!
! nrm   = integer. norm indicator. nrm = 1, means 1-norm, nrm =2
!                  means the 2-nrm, nrm = 0 means max norm
!
! a,
! ja,
! ia   = Matrix A in compressed sparse row format.
! 
! on return:
!----------
!
! diag = diagonal matrix stored as a vector containing the matrix
!        by which the columns have been scaled, i.e., on return 
!        we have B = A * Diag
!
! b, 
! jb, 
! ib	= resulting matrix B in compressed sparse row sparse format.
!
! ierr  = error message. ierr=0     : Normal return 
!                        ierr=i > 0 : Column number i is a zero row.
! Notes:
!-------
! 1)     The column dimension of A is not needed. 
! 2)     algorithm in place (B can take the place of A).
!-----------------------------------------------------------------
      call cnrms (nrow,nrm,a,ja,ia,diag)
      ierr = 0
      do 1 j=1, nrow
         if (diag(j) .eq. 0.0) then
            ierr = j 
            return
         else
            diag(j) = 1.0d0/diag(j)
         endif
 1    continue
      call amudia (nrow,job,a,ja,ia,diag,b,jb,ib)
      return
!--------end-of-coscal-------------------------------------------------- 
!-----------------------------------------------------------------------
      end
!-----------------------------------------------------------------------
      subroutine addblk(nrowa, ncola, a, ja, ia, ipos, jpos, job,         &
     & nrowb, ncolb, b, jb, ib, nrowc, ncolc, c, jc, ic, nzmx, ierr)
!      implicit none
      integer nrowa, nrowb, nrowc, ncola, ncolb, ncolc, ipos, jpos
      integer nzmx, ierr, job
      integer ja(1:*), ia(1:*), jb(1:*), ib(1:*), jc(1:*), ic(1:*)
      real*8 a(1:*), b(1:*), c(1:*)
!-----------------------------------------------------------------------
!     This subroutine adds a matrix B into a submatrix of A whose 
!     (1,1) element is located in the starting position (ipos, jpos). 
!     The resulting matrix is allowed to be larger than A (and B), 
!     and the resulting dimensions nrowc, ncolc will be redefined 
!     accordingly upon return.  
!     The input matrices are assumed to be sorted, i.e. in each row
!     the column indices appear in ascending order in the CSR format.
!-----------------------------------------------------------------------
! on entry:
! ---------
! nrowa    = number of rows in A.
! bcola    = number of columns in A.
! a,ja,ia  = Matrix A in compressed sparse row format with entries sorted
! nrowb    = number of rows in B.
! ncolb    = number of columns in B.
! b,jb,ib  = Matrix B in compressed sparse row format with entries sorted
!
! nzmax	   = integer. The  length of the arrays ! and jc. addblk will 
!            stop if the number of nonzero elements in the matrix C
!            exceeds nzmax. See ierr.
! 
! on return:
!----------
! nrowc    = number of rows in C.
! ncolc    = number of columns in C.
! !,jc,ic  = resulting matrix C in compressed sparse row sparse format
!            with entries sorted ascendly in each row. 
!	    
! ierr	   = integer. serving as error message. 
!         ierr = 0 means normal return,
!         ierr .gt. 0 means that addblk stopped while computing the
!         i-th row  of C with i=ierr, because the number 
!         of elements in C exceeds nzmax.
!
! Notes: 
!-------
!     this will not work if any of the two input matrices is not sorted
!-----------------------------------------------------------------------
      logical values
      integer i,j1,j2,ka,kb,kc,kamax,kbmax
      values = (job .ne. 0) 
      ierr = 0
      nrowc = max(nrowa, nrowb+ipos-1)
      ncolc = max(ncola, ncolb+jpos-1)
      kc = 1
      kbmax = 0
      ic(1) = kc
!
      do 10 i=1, nrowc
         if (i.le.nrowa) then
            ka = ia(i)
            kamax = ia(i+1)-1
         else
            ka = ia(nrowa+1)
         end if
         if ((i.ge.ipos).and.((i-ipos).le.nrowb)) then
            kb = ib(i-ipos+1)
            kbmax = ib(i-ipos+2)-1 
         else
            kb = ib(nrowb+1)
         end if
!
!     a do-while type loop -- goes through all the elements in a row.
!
 20      continue 
         if (ka .le. kamax) then
            j1 = ja(ka)
         else
            j1 = ncolc+1
         endif
         if (kb .le. kbmax) then 
            j2 = jb(kb) + jpos - 1
         else 
            j2 = ncolc+1
         endif
!
!     if there are more elements to be added.
!
         if ((ka .le. kamax .or. kb .le. kbmax) .and.                     &
     &        (j1 .le. ncolc .or. j2 .le. ncolc)) then
!
!     three cases
!
            if (j1 .eq. j2) then 
               if (values) c(kc) = a(ka)+b(kb)
               jc(kc) = j1
               ka = ka+1
               kb = kb+1
               kc = kc+1
            else if (j1 .lt. j2) then
               jc(kc) = j1
               if (values) c(kc) = a(ka)
               ka = ka+1
               kc = kc+1
            else if (j1 .gt. j2) then
               jc(kc) = j2
               if (values) c(kc) = b(kb)
               kb = kb+1
               kc = kc+1
            endif
            if (kc .gt. nzmx) goto 999
            goto 20
         end if
         ic(i+1) = kc
 10   continue
      return
 999  ierr = i 
      return
!---------end-of-addblk------------------------------------------------- 
      end
!----------------------------------------------------------------------- 
      subroutine get1up (n,ja,ia,ju)
      integer  n, ja(*),ia(*),ju(*)
!----------------------------------------------------------------------
! obtains the first element of each row of the upper triangular part
! of a matrix. Assumes that the matrix is already sorted.
!-----------------------------------------------------------------------
! parameters
! input
! ----- 
! ja      = integer array containing the column indices of aij
! ia      = pointer array. ia(j) contains the position of the 
!           beginning of row j in ja
! 
! output 
! ------ 
! ju      = integer array of length n. ju(i) is the address in ja 
!           of the first element of the uper triangular part of
!           of A (including rthe diagonal. Thus if row i does have
!           a nonzero diagonal element then ju(i) will point to it.
!           This is a more general version of diapos.
!-----------------------------------------------------------------------
! local vAriables
      integer i, k 
!     
      do 5 i=1, n
         ju(i) = 0
         k = ia(i) 
!
 1       continue
         if (ja(k) .ge. i) then
            ju(i) = k
            goto 5
         elseif (k .lt. ia(i+1) -1) then
            k=k+1
! 
! go try next element in row 
! 
            goto 1
         endif 
 5    continue
      return
!-----end-of-get1up-----------------------------------------------------
      end 
!----------------------------------------------------------------------
      subroutine xtrows (i1,i2,a,ja,ia,ao,jao,iao,iperm,job)
      integer i1,i2,ja(*),ia(*),jao(*),iao(*),iperm(*),job
      real*8 a(*),ao(*) 
!-----------------------------------------------------------------------
! this subroutine extracts given rows from a matrix in CSR format. 
! Specifically, rows number iperm(i1), iperm(i1+1), ...., iperm(i2)
! are extracted and put in the output matrix ao, jao, iao, in CSR
! format.  NOT in place. 
! Youcef Saad -- coded Feb 15, 1992. 
!-----------------------------------------------------------------------
! on entry:
!----------
! i1,i2   = two integers indicating the rows to be extracted.
!           xtrows will extract rows iperm(i1), iperm(i1+1),..,iperm(i2),
!           from original matrix and stack them in output matrix 
!           ao, jao, iao in csr format
!
! a, ja, ia = input matrix in csr format
!
! iperm	= integer array of length nrow containing the reverse permutation 
!         array for the rows. row number iperm(j) in permuted matrix PA
!         used to be row number j in unpermuted matrix.
!         ---> a(i,j) in the permuted matrix was a(iperm(i),j) 
!         in the inout matrix.
!
! job	= integer indicating the work to be done:
! 		job .ne. 1 : get structure only of output matrix,,
!               i.e., ignore real values. (in which case arrays a 
!               and ao are not used nor accessed).
! 		job = 1	get complete data structure of output matrix. 
!               (i.e., including arrays ao and iao).
!------------
! on return: 
!------------ 
! ao, jao, iao = input matrix in a, ja, ia format
! note : 
!        if (job.ne.1)  then the arrays a and ao are not used.
!----------------------------------------------------------------------!
!           Y. Saad, revised May  2, 1990                              !
!----------------------------------------------------------------------!
      logical values
      values = (job .eq. 1) 
!
! copying 
!
      ko = 1
      iao(1) = ko
      do 100 j=i1,i2 
!
! ii=iperm(j) is the index of old row to be copied.
!        
         ii = iperm(j) 
         do 60 k=ia(ii), ia(ii+1)-1 
            jao(ko) = ja(k) 
            if (values) ao(ko) = a(k)
            ko = ko+1
 60      continue
         iao(j-i1+2) = ko
 100  continue
!
      return
!---------end-of-xtrows------------------------------------------------- 
!-----------------------------------------------------------------------
      end
!-----------------------------------------------------------------------
      subroutine csrkvstr(n, ia, ja, nr, kvstr)
!-----------------------------------------------------------------------
      integer n, ia(n+1), ja(*), nr, kvstr(*)
!-----------------------------------------------------------------------
!     Finds block row partitioning of matrix in CSR format.
!-----------------------------------------------------------------------
!     On entry:
!--------------
!     n       = number of matrix scalar rows
!     ia,ja   = input matrix sparsity structure in CSR format
!
!     On return:
!---------------
!     nr      = number of block rows
!     kvstr   = first row number for each block row
!
!     Notes:
!-----------
!     Assumes that the matrix is sorted by columns.
!     This routine does not need any workspace.
!
!-----------------------------------------------------------------------
!     local variables
      integer i, j, jdiff
!-----------------------------------------------------------------------
      nr = 1
      kvstr(1) = 1
!---------------------------------
      do i = 2, n
         jdiff = ia(i+1)-ia(i)
         if (jdiff .eq. ia(i)-ia(i-1)) then
            do j = ia(i), ia(i+1)-1
               if (ja(j) .ne. ja(j-jdiff)) then
                  nr = nr + 1
                  kvstr(nr) = i
                  goto 299
               endif
            enddo
 299        continue
         else
 300        nr = nr + 1
            kvstr(nr) = i
         endif
      enddo
      kvstr(nr+1) = n+1
!---------------------------------
      return
      end
!-----------------------------------------------------------------------
!------------------------end-of-csrkvstr--------------------------------
      subroutine csrkvstc(n, ia, ja, nc, kvstc, iwk)
!-----------------------------------------------------------------------
      integer n, ia(n+1), ja(*), nc, kvstc(*), iwk(*)
!-----------------------------------------------------------------------
!     Finds block column partitioning of matrix in CSR format.
!-----------------------------------------------------------------------
!     On entry:
!--------------
!     n       = number of matrix scalar rows
!     ia,ja   = input matrix sparsity structure in CSR format
!
!     On return:
!---------------
!     nc      = number of block columns
!     kvstc   = first column number for each block column
!
!     Work space:
!----------------
!     iwk(*) of size equal to the number of scalar columns plus one.
!        Assumed initialized to 0, and left initialized on return.
!
!     Notes:
!-----------
!     Assumes that the matrix is sorted by columns.
!
!-----------------------------------------------------------------------
!     local variables
      integer i, j, k, ncol
!
!-----------------------------------------------------------------------
!-----use ncol to find maximum scalar column number
      ncol = 0
!-----mark the beginning position of the blocks in iwk
      do i = 1, n
         if (ia(i) .lt. ia(i+1)) then
            j = ja(ia(i))
            iwk(j) = 1
            do k = ia(i)+1, ia(i+1)-1
               j = ja(k)
               if (ja(k-1).ne.j-1) then
                  iwk(j) = 1
                  iwk(ja(k-1)+1) = 1
               endif
            enddo
            iwk(j+1) = 1
            ncol = max0(ncol, j)
         endif
      enddo
!---------------------------------
      nc = 1
      kvstc(1) = 1
      do i = 2, ncol+1
         if (iwk(i).ne.0) then
            nc = nc + 1
            kvstc(nc) = i
            iwk(i) = 0
         endif
      enddo
      nc = nc - 1
!---------------------------------
      return
      end
!-----------------------------------------------------------------------
!------------------------end-of-csrkvstc--------------------------------
!-----------------------------------------------------------------------
      subroutine kvstmerge(nr, kvstr, nc, kvstc, n, kvst)
!-----------------------------------------------------------------------
      integer nr, kvstr(nr+1), nc, kvstc(nc+1), n, kvst(*)
!-----------------------------------------------------------------------
!     Merges block partitionings, for conformal row/col pattern.
!-----------------------------------------------------------------------
!     On entry:
!--------------
!     nr,nc   = matrix block row and block column dimension
!     kvstr   = first row number for each block row
!     kvstc   = first column number for each block column
!
!     On return:
!---------------
!     n       = conformal row/col matrix block dimension
!     kvst    = conformal row/col block partitioning
!
!     Notes:
!-----------
!     If matrix is not square, this routine returns without warning.
!
!-----------------------------------------------------------------------
!-----local variables
      integer i,j
!---------------------------------
      if (kvstr(nr+1) .ne. kvstc(nc+1)) return
      i = 1
      j = 1
      n = 1
  200 if (i .gt. nr+1) then
         kvst(n) = kvstc(j)
         j = j + 1
      elseif (j .gt. nc+1) then
         kvst(n) = kvstr(i)
         i = i + 1
      elseif (kvstc(j) .eq. kvstr(i)) then
         kvst(n) = kvstc(j)
         j = j + 1
         i = i + 1
      elseif (kvstc(j) .lt. kvstr(i)) then
         kvst(n) = kvstc(j)
         j = j + 1
      else
         kvst(n) = kvstr(i)
         i = i + 1
      endif
      n = n + 1
      if (i.le.nr+1 .or. j.le.nc+1) goto 200
      n = n - 2
!---------------------------------
      return
!------------------------end-of-kvstmerge-------------------------------
      end
